# -*- coding: utf-8 -*-
# :Project:   hurm
# :Created:   mar 22 dic 2015 16:04:16 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016 Lele Gaifax
#

from .activities import activities
from .availabilities import availabilities
from .duties import duties
from .editions import editions
from .locations import locations
from .persons import persons, password_manager
from .personactivities import personactivities
from .tasks import tasks



# -*- coding: utf-8 -*-
# :Project:   hurm -- SQLAlchemy declaration of table availabilities
# :Created:   mer 23 dic 2015 11:13:05 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016 Lele Gaifax
#

import sqlalchemy as sa

from .. import metadata, translatable_string as _
from . import domains


availabilities = sa.Table(
    'availabilities', metadata,

    sa.Column(
        'idavailability', domains.integer_t, sa.Sequence('gen_idavailability', optional=True),
        nullable=False,
        primary_key=True,
        info={'label': _("Availability ID"),
              'hint': _("Unique ID of the availability")}),

    sa.Column(
        'idedition', domains.integer_t, sa.ForeignKey('editions.idedition'),
        nullable=False,
        info={'label': _("Edition ID"),
              'hint': _("ID of the edition")}),

    sa.Column(
        'idperson', domains.integer_t, sa.ForeignKey('persons.idperson'),
        nullable=False,
        info={'label': _("Person ID"),
              'hint': _("ID of the related person")}),

    sa.Column(
        'date', domains.date_t,
        nullable=False,
        info={'label': _("Date"),
              'hint': _("Reference date")}),

    sa.Column(
        'starttime', domains.time_t,
        info={'label': _("Start"),
              'hint': _("Starting time of the availability,"
                        " empty to mean from start of the day")}),

    sa.Column(
        'endtime', domains.time_t,
        info={'label': _("End"),
              'hint': _("Ending time of the availability,"
                        " empty to mean till the end of the day")}),

    sa.Column(
        'note', domains.text_t,
        info={'label': _("Note"),
              'hint': _("Note")}),
)



# -*- coding: utf-8 -*-
# :Project:   hurm
# :Created:   ven 22 gen 2016 16:53:21 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from datetime import date
from os import getenv

import pytest

from sqlalchemy import create_engine

from hurm.db import metadata, entities


@pytest.fixture(scope="session", autouse=True)
def engine():
    metadata.bind = entities.DBSession.bind = create_engine(getenv('TEST_DB_URL'))


@pytest.fixture(scope="module")
def session():
    return entities.DBSession()


@pytest.fixture
def edition_test(session):
    return session.query(entities.Edition).filter_by(description='Test edition').one()


@pytest.fixture
def john_doe(session):
    return session.query(entities.Person).filter_by(lastname='Doe').one()


@pytest.fixture
def jane_tree(session):
    return session.query(entities.Person).filter_by(lastname='Tree').one()


@pytest.fixture
def bob_quartz(session):
    return session.query(entities.Person).filter_by(lastname='Quartz').one()


@pytest.fixture
def presence(session):
    return session.query(entities.Activity).filter_by(description='Presence').one()


@pytest.fixture
def supervision(session):
    return session.query(entities.Activity).filter_by(description='Supervision').one()


@pytest.fixture
def reception(session):
    return session.query(entities.Location).filter_by(description='Reception').one()


@pytest.fixture
def reception_presence_05_01(session, edition_test, presence, reception):
    return session.query(entities.Task).filter_by(idedition=edition_test.idedition,
                                                  idactivity=presence.idactivity,
                                                  idlocation=reception.idlocation,
                                                  date=date(2016, 5, 1)).one()


@pytest.fixture
def availability_john_doe_04_30(session, edition_test, john_doe):
    return session.query(entities.Availability).filter_by(idedition=edition_test.idedition,
                                                          idperson=john_doe.idperson,
                                                          date=date(2016, 4, 30)).one()


@pytest.fixture
def availability_john_doe_05_01(session, edition_test, john_doe):
    return session.query(entities.Availability).filter_by(idedition=edition_test.idedition,
                                                          idperson=john_doe.idperson,
                                                          date=date(2016, 5, 1)).one()



# -*- coding: utf-8 -*-
# :Project:   hurm -- Abstract base class for entities
# :Created:   sab 02 gen 2016 15:24:29 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import logging

import sqlalchemy as sa
from sqlalchemy import orm

from .. import tables


DBSession = orm.scoped_session(orm.sessionmaker())
"The global SA session maker"

logger = logging.getLogger(__name__)


class AbstractBase(object):
    "Abstract base entity class."

    def __init__(self, **kwargs):
        """Initialize an instance of this class.

        :param kwargs: initial values for the instance attributes

        Initialize a new instance of this class, copying any keyword
        argument into the corresponding instance attribute.
        """

        cls_ = type(self)
        for k in kwargs:
            if not hasattr(cls_, k):
                raise TypeError(
                    "%r is an invalid keyword argument for %s" %
                    (k, cls_.__name__))
            setattr(self, k, kwargs[k])

    def delete(self):
        "Delete this instance from the database."

        sa.orm.object_session(self).delete(self)

    def update(self, data, missing_only=False):
        """Update the instance with given data.

        :param data: a mapping kind of container
        :param missing_only: a boolean flag, ``False`` by default
        :rtype: dict
        :returns: a mapping between field name and a tuple ``(oldvalue, newvalue)``,
                  for each modified field

        If `missing_only` is ``True`` then only the fields that are currently `empty`
        (that is, their value is either ``None`` or an empty string) are updated.
        """

        changes = {}

        for attr in data:
            if hasattr(self, attr):
                cvalue = getattr(self, attr)
                if missing_only:
                    if not (cvalue is None or cvalue == '' or
                            cvalue is True or cvalue is False):
                        continue
                nvalue = data[attr]
                if cvalue != nvalue:
                    setattr(self, attr, nvalue)
                    changes[attr] = (cvalue, nvalue)
            else:
                logger.warning("Ignoring change to attribute %r for %r", attr, self)

        return changes

    def __repr__(self):
        "Return an ASCII representation of the entity."

        from itertools import zip_longest

        mapper = sa.orm.object_mapper(self)
        pkeyf = mapper.primary_key
        try:
            pkeyv = mapper.primary_key_from_instance(self)
        except sa.orm.exc.DetachedInstanceError:
            keys = u"detached-instance"
        else:
            keys = u', '.join(u'%s=%s' % (f.name, v)
                              for f, v in zip_longest(pkeyf, pkeyv))
        return u'<%s %s>' % (# pragma: no cover
                             self.__class__.__name__, keys)


from .activity import Activity
from .availability import Availability
from .duty import Duty
from .edition import Edition
from .location import Location
from .person import Person
from .personactivity import PersonActivity
from .task import Task


orm.mapper(Activity, tables.activities, properties={
})

orm.mapper(Availability, tables.availabilities, properties={
    'edition': orm.relationship(Edition),
})

orm.mapper(Duty, tables.duties, properties={
    'task_date': sa.orm.deferred(sa.select([tables.tasks.c.date])
                                 .where(tables.tasks.c.idtask == tables.duties.c.idtask)),
})

orm.mapper(Person, tables.persons, properties={
    'activities': orm.relationship(Activity, secondary=tables.personactivities),
    'availabilities': orm.relationship(Availability,
                                       backref='person',
                                       order_by=[Availability.date, Availability.starttime]),
    'duties': orm.relationship(Duty,
                               backref='person',
                               order_by=[Duty.task_date, Duty.starttime]),
})

orm.mapper(Task, tables.tasks, properties={
    'activity': orm.relationship(Activity),
    'duties': orm.relationship(Duty,
                               backref='task',
                               order_by=[Duty.starttime]),
    'location': orm.relationship(Location)
})

orm.mapper(Edition, tables.editions, properties={
    'tasks': orm.relationship(Task,
                              backref='edition',
                              order_by=[Task.date, Task.starttime]),
})

orm.mapper(Location, tables.locations, properties={
})

orm.mapper(PersonActivity, tables.personactivities, properties={
    'activity': orm.relationship(Activity),
    'person': orm.relationship(Person, backref='preferred_activities'),
})



# -*- coding: utf-8 -*-
# :Project:   hurm
# :Created:   sab 23 gen 2016 09:55:21 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from datetime import time

from sqlalchemy import exc as saexc

from hurm.db import entities


def test_out_task_duty(session, reception_presence_05_01, bob_quartz):
    d = entities.Duty()
    d.person = bob_quartz
    d.task = reception_presence_05_01
    d.starttime = time(7, 0)
    d.endtime = time(8, 0)
    session.add(d)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'duty time outside task time' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()


def test_out_person_duty(session, reception_presence_05_01, john_doe):
    d = entities.Duty()
    d.person = john_doe
    d.task = reception_presence_05_01
    d.starttime = time(15, 0)
    d.endtime = time(16, 0)
    session.add(d)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'duty time outside person availability' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()


def test_non_overlapping_duty(session, reception_presence_05_01, bob_quartz):
    d = entities.Duty()
    d.person = bob_quartz
    d.task = reception_presence_05_01
    d.starttime = time(8, 0)
    d.endtime = time(10, 0)
    session.add(d)
    try:
        session.flush()
    finally:
        session.rollback()


def test_overlapping_duty(session, reception_presence_05_01, jane_tree):
    d = entities.Duty()
    d.person = jane_tree
    d.task = reception_presence_05_01
    d.starttime = time(11, 0)
    d.endtime = time(11, 30)
    session.add(d)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'overlapped duty' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()



# -*- coding: utf-8 -*-
# :Project:   hurm -- Tests for the helper functions
# :Created:   mar 09 feb 2016 16:56:54 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from datetime import date

from hurm.db import entities, tables, utils


def test_normalize():
    assert utils.normalize(None) is None
    assert utils.normalize('lele gaifax') == 'Lele Gaifax'
    assert utils.normalize('LELE') == 'Lele'
    assert utils.normalize('LeLe', title=False) == 'LeLe'


def test_njoin():
    assert utils.njoin([1,2,3]) == '1, 2 and 3'
    assert utils.njoin([1,2]) == '1 and 2'
    assert utils.njoin([1]) == '1'
    assert utils.njoin([]) == ''
    assert utils.njoin([1,2], stringify=lambda x: str(x*10)) == '10 and 20'
    assert utils.njoin(['first', None, False, '', 'last']) == 'first and last'
    assert utils.njoin([1,0,2]) == '1, 0 and 2'


def test_changes_summary():
    assert utils.changes_summary(dict(a=(None, 1))) == 'changed a to 1'
    assert utils.changes_summary(dict(a=(False, True))) == 'changed a from False to True'
    assert utils.changes_summary(dict(a=(0,1), b=('foo','bar'))) == \
        'changed a from 0 to 1 and b from "foo" to "bar"'
    assert utils.changes_summary(dict(a=(0,1), b=(None,'bar'), c=(True,None))) == \
        'changed a from 0 to 1, b to "bar" and c from True to None'


def test_table_from_primary_key(session):
    assert utils.table_from_primary_key('idedition') is tables.editions
    try:
        utils.table_from_primary_key('foobar')
    except RuntimeError:
        pass
    else:
        assert False, "Should raise a RuntimeError"


def test_entity_from_primary_key(session):
    assert utils.entity_from_primary_key('idedition') is entities.Edition
    try:
        utils.entity_from_primary_key('foobar')
    except RuntimeError:
        pass
    else:
        assert False, "Should raise a RuntimeError"


def test_save_changes(session, edition_test, presence, supervision):
    try:
        i, m, d = utils.save_changes(session, None,
                                     [('idedition', {'idedition': edition_test.idedition,
                                                     'description': 'this is a test',
                                                     'does_not_exist': 'dummy'}),
                                      ('idperson', {'firstname': 'foo',
                                                    'lastname': 'bar',
                                                    'password': 'simple',
                                                    'role': '',
                                                    'idactivities': [presence.idactivity]}),
                                     ('idavailability', {'idedition': edition_test.idedition,
                                                         'idperson': 0,
                                                         'date': date(2016, 4, 30),
                                                         'note': 'dummy note'})],
                                     [])
        assert list(i[0].keys()) == ['idperson']
        assert m == [{'idedition': edition_test.idedition}]
        assert d == []
        assert edition_test.description == 'this is a test'

        foobar_id = i[0]['idperson']
        foobar = session.query(entities.Person).get(foobar_id)
        assert foobar.firstname == 'Foo'
        assert foobar.lastname == 'Bar'
        assert foobar.password and tables.password_manager.check(foobar.password, 'simple')
        assert foobar.role is None
        assert foobar.availabilities[0].note == 'dummy note'
        assert len(foobar.activities) == 1
        assert foobar.activities[0] is presence

        i, m, d = utils.save_changes(session, None,
                                     [('idperson', {'idperson': foobar_id,
                                                    'password': 'sample',
                                                    'idactivities': [supervision.idactivity]})],
                                     [])

        foobar = session.query(entities.Person).get(foobar_id)
        assert foobar.password and tables.password_manager.check(foobar.password, 'sample')
        assert len(foobar.activities) == 1
        assert foobar.activities[0] is supervision

        i, m, d = utils.save_changes(session, None,
                                     [('idperson', {'idperson': foobar_id,
                                                    'idactivities': None})],
                                     [])

        foobar = session.query(entities.Person).get(foobar_id)
        assert len(foobar.activities) == 0

        idavail = foobar.availabilities[0].idavailability
        i, m, d = utils.save_changes(session, None, [],
                                     [('idavailability', idavail)])
        assert i == []
        assert m == []
        assert d == [{'idavailability': idavail}]

        idpers = foobar.idperson
        i, m, d = utils.save_changes(session, None, [],
                                     [('idperson', idpers)])
        assert i == []
        assert m == []
        assert d == [{'idperson': idpers}]
    finally:
        session.rollback()


def test_dumpy(edition_test):
    dump = utils.dump(edition_test, [{
        'entity': 'hurm.db.entities.edition.Edition',
        'key': 'description',
        'other': 'startdate,enddate,tasks'.split(',')
    }, {
        'entity': 'hurm.db.entities.activity.Activity',
        'key': 'description'
    }, {
        'entity': 'hurm.db.entities.location.Location',
        'key': 'description',
        'other': 'address,city,country,province,zip'.split(',')
    }, {
        'entity': 'hurm.db.entities.task.Task',
        'key': 'edition,activity,location,date,starttime'.split(','),
        'other': 'endtime,npersons,duties'.split(',')
    }, {
        'entity': 'hurm.db.entities.person.Person',
        'key': 'lastname,firstname'.split(','),
        'other': 'birthdate,mobile,phone,email,role,availabilities,preferred_activities'.split(',')
    }, {
        'entity': 'hurm.db.entities.personactivity.PersonActivity',
        'key': 'person,activity'.split(','),
    }, {
        'entity': 'hurm.db.entities.availability.Availability',
        'key': 'edition,person,date,starttime'.split(','),
        'other': 'endtime'
    }, {
        'entity': 'hurm.db.entities.duty.Duty',
        'key': 'person,task,starttime'.split(','),
        'other': 'endtime'
    }])

    assert dump[0]['data'][0]['description'] == edition_test.description



# -*- coding: utf-8 -*-
# :Project:   hurm -- SQLAlchemy declaration of table personactivities
# :Created:   gio 18 feb 2016 12:31:09 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import sqlalchemy as sa

from .. import metadata, translatable_string as _
from . import domains


personactivities = sa.Table(
    'personactivities', metadata,

    sa.Column(
        'idperson', domains.integer_t, sa.ForeignKey('persons.idperson'),
        nullable=False,
        primary_key=True,
        info={'label': _("Person ID"),
              'hint': _("ID of the related person")}),

    sa.Column(
        'idactivity', domains.integer_t, sa.ForeignKey('activities.idactivity'),
        nullable=False,
        primary_key=True,
        info={'label': _("Activity ID"),
              'hint': _("ID of the related activity")}),
)



# -*- coding: utf-8 -*-
# :Project:   hurm -- Database modelization
# :Created:   lun 14 dic 2015 15:56:04 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016 Lele Gaifax
#

import logging

from sqlalchemy import MetaData
from translationstring import TranslationStringFactory


metadata = MetaData()
"Container of all tables meta information."


translatable_string = TranslationStringFactory('hurm-db')
"A function to make a translatable string."


logger = logging.getLogger(__name__)
"Logger used for the DB related messages"


from .tables import *



# -*- coding: utf-8 -*-
# :Project:   hurm -- Activity class, mapped to table activities
# :Created:   sab 02 gen 2016 15:32:31 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from . import AbstractBase


class Activity(AbstractBase):
    pass



# -*- coding: utf-8 -*-
# :Project:   hurm -- SQLAlchemy declaration of table locations
# :Created:   mar 22 dic 2015 18:17:59 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016 Lele Gaifax
#

import sqlalchemy as sa

from .. import metadata, translatable_string as _
from . import domains


locations = sa.Table(
    'locations', metadata,

    sa.Column(
        'idlocation', domains.integer_t, sa.Sequence('gen_idlocation', optional=True),
        nullable=False,
        primary_key=True,
        info={'label': _("Location ID"),
              'hint': _("Unique ID of the location")}),

    sa.Column(
        'description', domains.description_t,
        nullable=False,
        unique=True,
        info={'label': _("Location"),
              'hint': _("Description of the location")}),

    sa.Column(
        'address', domains.description_t,
        info=dict(
            label=_("Address"),
            hint=_("Street address"))),

    sa.Column(
        'city', domains.description_t,
        info=dict(
            label=_("City"),
            hint=_("City name"))),

    sa.Column(
        'province', domains.province_t,
        info=dict(
            label=_("Province code"),
            hint=_("Code of the province"))),

    sa.Column(
        'zip', domains.shortcode_t,
        info=dict(
            label=_("Zip code"),
            hint=_("Zip code of the city"))),

    sa.Column(
        'country', domains.country_t,
        info=dict(
            label=_("Country code"),
            hint=_("Code of the country"))),

    sa.Column(
        'latitude', domains.latlng_t,
        info={'label': _("Latitude"),
              'hint': _("Latitude of the location")}),

    sa.Column(
        'longitude', domains.latlng_t,
        info={'label': _("Longitude"),
              'hint': _("Longitude of the location")}),

    sa.Column(
        'phone', domains.phone_t,
        info={'label': _("Phone"),
              'hint': _("Phone number of the location")}),

    sa.Column(
        'mobile', domains.phone_t,
        info={'label': _("Mobile"),
              'hint': _("Mobile number of the location")}),

    sa.Column(
        'email', domains.email_t,
        info={'label': _("Email"),
              'hint': _("Email address of the location")}),

    sa.Column(
        'note', domains.text_t,
        info={'label': _("Note"),
              'hint': _("Note")}),
)



# -*- coding: utf-8 -*-
# :Project:   hurm -- PersonActivity class, mapped to table personactivities
# :Created:   gio 18 feb 2016 12:35:11 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from . import AbstractBase


class PersonActivity(AbstractBase):
    pass



# -*- coding: utf-8 -*-
# :Project:   hurm
# :Created:   ven 22 gen 2016 21:37:30 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from datetime import date, time

from sqlalchemy import exc as saexc, sql

from hurm.db import entities


def test_out_edition_task(session, edition_test, presence, reception):
    t = entities.Task()
    t.edition = edition_test
    t.location = reception
    t.activity = presence
    t.date = date(2015, 5, 1)
    t.starttime = time(14, 0)
    t.endtime = time(16, 0)
    t.npersons = 1
    session.add(t)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'task date outside allowed period' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()


def test_non_overlapping_task(session, edition_test, presence, reception):
    t = entities.Task()
    t.edition = edition_test
    t.location = reception
    t.activity = presence
    t.date = date(2016, 5, 1)
    t.starttime = time(22, 0)
    t.endtime = time(23, 0)
    t.npersons = 1
    session.add(t)
    try:
        session.flush()
    finally:
        session.rollback()


def test_overlapping_task(session, edition_test, presence, reception):
    t = entities.Task()
    t.edition = edition_test
    t.location = reception
    t.activity = presence
    t.date = date(2016, 5, 1)
    t.starttime = time(14, 0)
    t.endtime = time(16, 0)
    t.npersons = 1
    session.add(t)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'overlapped task' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()


def test_invalid_task_edit(session, reception_presence_05_01):
    reception_presence_05_01.endtime = time(9, 0)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'duties outside task time range' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()


def test_task_timeline(session, reception_presence_05_01):
    c = session.connection()
    q = sql.text("SELECT starttime, endtime, npersons FROM task_timeline(:idtask)")
    r = c.execute(q, idtask=reception_presence_05_01.idtask).fetchall()
    assert len(r) == 4
    assert r[0]['starttime'] == reception_presence_05_01.starttime
    assert r[-1]['endtime'] == reception_presence_05_01.endtime



# -*- coding: utf-8 -*-
# :Project:   hurm -- Data domains
# :Created:   lun 14 dic 2015 16:00:45 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016 Lele Gaifax
#

from sqlalchemy import (
    CHAR,
    Date,
    Integer,
    Numeric,
    Time,
    Unicode,
    UnicodeText,
    VARCHAR,
    )
from sqlalchemy.types import TypeDecorator

from ..utils import normalize


class Description(TypeDecorator):
    impl = Unicode

    def process_bind_param(self, value, dialect):
        return normalize(value)


class Name(TypeDecorator):
    impl = Unicode

    def process_bind_param(self, value, dialect):
        return normalize(value, True)


country_t = CHAR(2)
"A country ISO 3166 alpha2 code"

date_t = Date()
"A date"

description_t = Description(100)
"A long description"

email_t = VARCHAR(50)
"An email address"

integer_t = Integer()
"An integer value"

latlng_t = Numeric(10,6)
"A latitude or a longitude"

name_t = Name(50)
"A fifty characters long name"

password_t = VARCHAR(60)
"A password hash"

phone_t = VARCHAR(20)
"A phone number"

province_t = VARCHAR(6)
"A province ISO 3166-2 code"

shortcode_t = VARCHAR(10)
"A short code"

text_t = UnicodeText()
"An arbitrarily long description."

time_t = Time()
"A time of the day"



# -*- coding: utf-8 -*-
# :Project:   hurm -- Edition class, mapped to table editions
# :Created:   sab 02 gen 2016 15:34:14 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from . import AbstractBase


class Edition(AbstractBase):
    pass



# -*- coding: utf-8 -*-
# :Project:   hurm
# :Created:   mer 30 dic 2015 18:17:03 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016 Lele Gaifax
#

import atexit, datetime, os, re, subprocess, tempfile

import yaml


def time_constructor(loader, node):
    value = loader.construct_scalar(node)
    if value.startswith('T'):
        value = value[1:]
    parts = map(int, value.split(':'))
    return datetime.time(*parts)

yaml.add_constructor('!time', time_constructor)
yaml.add_implicit_resolver('!time', re.compile(r'^T?\d{2}:\d{2}(:\d{2})?$'), ['T'])


def find_in_path(executable):
    path = os.getenv('PATH', os.defpath)
    for d in path.split(os.pathsep):
        epath = os.path.abspath(os.path.join(d, executable))
        if os.path.exists(epath):
            return epath
    return None

def decipher(fname):
    gpg = find_in_path('gpg2') or find_in_path('gpg')
    if not gpg:
        print("ERROR: GPG command line tool is missing!")
        return fname

    print("Input file %s is encrypted, please enter passphrase" % fname)
    with tempfile.NamedTemporaryFile(suffix='.yaml') as f:
        tmpfname = f.name
    subprocess.run([gpg, '-q', '-o', tmpfname, fname], check=True)
    atexit.register(lambda n=tmpfname: os.unlink(n))
    return tmpfname

try:
    fnames
except NameError:
    # ignore, by any chance executed by the test runner
    fnames = []

fnames = [decipher(fname) if fname.endswith('.gpg') else fname for fname in fnames]



# -*- coding: utf-8 -*-
# :Project:   hurm
# :Created:   sab 23 gen 2016 19:30:39 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from datetime import date

from sqlalchemy import exc as saexc


def test_edition_repr(session, edition_test):
    assert 'Edition ' in repr(edition_test)


def test_invalid_edition_period(session, edition_test):
    edition_test.startdate = date(1968, 3, 18)
    edition_test.enddate = date(1968, 3, 28)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'availabilities outside edition period' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()

    edition_test.enddate = date(2016, 5, 1)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'tasks outside edition period' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()



# -*- coding: utf-8 -*-
# :Project:   hurm -- Person class, mapped to table persons
# :Created:   sab 02 gen 2016 15:34:36 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from sqlalchemy.orm import object_session

from .. import translatable_string as N_
from . import AbstractBase, DBSession

from ..tables import password_manager


class Person(AbstractBase):
    def __init__(self, **kwargs):
        if 'password' in kwargs and kwargs['password']:
            kwargs['password'] = password_manager.encode(kwargs['password'])

        if 'idactivities' in kwargs:
            self._update_idactivities(kwargs.pop('idactivities'))

        super().__init__(**kwargs)

    def update(self, data, **kwargs):
        if 'password' in data and data['password']:
            data['password'] = password_manager.encode(data['password'])

        if 'idactivities' in data:
            oldidas, newidas = self._update_idactivities(data.pop('idactivities'))
        else:
            oldidas = None

        changes = super().update(data, **kwargs)

        if oldidas is not None:
            changes['idactivities'] = (oldidas, newidas)

        return changes

    def _update_idactivities(self, ids):
        from .activity import Activity

        if isinstance(ids, str):
            ids = list(map(int, ids.split(',')))

        oldidas = [activity.idactivity for activity in self.activities]

        to_be_removed = []
        for activity in self.activities:
            if ids and activity.idactivity in ids:
                ids.remove(activity.idactivity)
            else:
                to_be_removed.append(activity)

        for activity in to_be_removed:
            self.activities.remove(activity)

        if ids:
            session = object_session(self) or DBSession()
            for id in ids:
                self.activities.append(session.query(Activity).get(id))

        newidas = [activity.idactivity for activity in self.activities]

        if set(oldidas) == set(newidas):
            oldidas = newidas = None

        return oldidas, newidas

    @property
    def fullname(self):
        # TRANSLATORS: this is the full name of a person
        return N_('$firstname $lastname',
                  mapping=dict(firstname=self.firstname,
                               lastname=self.lastname))

    @property
    def abbreviated_fullname(self):
        # TRANSLATORS: this is the abbreviated full name of a person
        return N_('$firstname_initial. $lastname',
                  mapping=dict(firstname_initial=self.firstname[0],
                               lastname=self.lastname))



# -*- coding: utf-8 -*-
# :Project:   hurm -- SQLAlchemy declaration of table persons
# :Created:   mar 22 dic 2015 18:45:01 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016 Lele Gaifax
#

import sqlalchemy as sa

from cryptacular.bcrypt import BCRYPTPasswordManager

from .. import metadata, translatable_string as _
from . import domains


password_manager = BCRYPTPasswordManager()


persons = sa.Table(
    'persons', metadata,

    sa.Column(
        'idperson', domains.integer_t, sa.Sequence('gen_idperson', optional=True),
        nullable=False,
        primary_key=True,
        info={'label': _("Person ID"),
              'hint': _("Unique ID of the person")}),

    sa.Column(
        'firstname', domains.name_t,
        nullable=False,
        info={'label': _("First name"),
              'hint': _("First name of the person")}),

    sa.Column(
        'lastname', domains.name_t,
        nullable=False,
        info={'label': _("Last name"),
              'hint': _("Last name of the person")}),

    sa.Column(
        'role', domains.description_t,
        info={'label': _("Role"),
              'hint': _("Role of the person")}),

    sa.Column(
        'birthdate', domains.date_t,
        info={'label': _("Birthdate"),
              'hint': _("Date of birth")}),

    sa.Column(
        'phone', domains.phone_t,
        info={'label': _("Phone"),
              'hint': _("Phone number of the person")}),

    sa.Column(
        'mobile', domains.phone_t,
        info={'label': _("Mobile"),
              'hint': _("Mobile number of the person")}),

    sa.Column(
        'email', domains.email_t,
        unique=True,
        info={'label': _("Email"),
              'hint': _("Email address of the person")}),

    sa.Column(
        'password', domains.password_t,
        info={'label': _("Password"),
              'hint': _("Hashed password")}),

    sa.Column(
        'note', domains.text_t,
        info={'label': _("Note"),
              'hint': _("Note")}),
)



# -*- coding: utf-8 -*-
# :Project:   hurm -- Utility functions
# :Created:   mar 22 dic 2015 16:34:23 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016 Lele Gaifax
#

from sqlalchemy import inspect
from sqlalchemy.exc import NoInspectionAvailable, SQLAlchemyError
from sqlalchemy.orm.collections import InstrumentedList

from . import translatable_string as _, logger


def normalize(s, title=None):
    """Normalize the case of a string, removing spurious spaces.

    :param s: a string
    :param title: if `True` always titleize the string, if `False`
                  never do that, if `None` (default) only when the
                  input string is all lower case or all upper case
    :rtype: unicode

    ::

      >>> assert normalize(None) is None
      >>> print(normalize('lele gaifax'))
      Lele Gaifax
      >>> print(normalize('LELE'))
      Lele
      >>> print(normalize('LeLe', title=False))
      LeLe
    """

    if s is None:
        return None
    else:
        n = ' '.join(s.strip().split())
        if title != False and (title == True or
                               n == n.upper() or
                               n == n.lower()):
            n = n.title()
        return n


def njoin(elts, stringify=str, localized=True):
    """Given a sequence of items, concatenate them in a nice way.

    :param elts: a sequence of elements
    :param stringify: the stringification function applied to all elements,
                      by default :py:func:`str`
    :param localized: a boolean flag to disable the translation of the final 'and'
    :rtype: unicode

    If `elts` is empty returns an empty string; if it contains a single element, returns the
    stringified element; otherwise returns a string composed by all but the last elements
    stringified and joined by a comma, followed by the localized version of `and` followed by
    the last element stringified::

      >>> print(njoin([1,2,3]))
      1, 2 and 3
      >>> print(njoin([1,2]))
      1 and 2
      >>> print(njoin([1]))
      1
      >>> assert njoin([]) == ''
      >>> print(njoin([1,2], stringify=lambda x: str(x*10)))
      10 and 20

    Note that *falsey* elements are skipped::

      >>> print(njoin(['first', None, False, '', 'last']))
      first and last

    but ``0`` (*zero*) isn't considered a *falsey* value::

      >>> print(njoin([1,0,2]))
      1, 0 and 2
    """

    elts = [stringify(e) for e in elts if e or (e == 0 and e is not False)]
    if not elts:
        return ''
    elif len(elts) == 1:
        return elts[0]
    else:
        last = elts[-1]
        if localized:
            # TRANSLATORS: this is the final "conjunction" used when joining multiple
            # statements, for example "x, y and z".
            and_ = ' %s ' % _('and')
        else:
            and_ = ' and '
        return ', '.join(elts[:-1]) + and_ + last


def table_from_primary_key(pkname):
    """Given the name of a primary key, return the related table.

    :param pkname: the name of a primary key
    :rtype: a SQLAlchemy table
    """

    from . import metadata

    for t in metadata.sorted_tables:
        if len(t.primary_key.columns) == 1 and pkname in t.primary_key.columns:
            return t

    raise RuntimeError('Unknown PK: %s' % pkname)


def entity_from_primary_key(pkname):
    """Given the name of a primary key, return the mapped entity.

    :param pkname: the name of a primary key
    :rtype: a mapped class
    """

    from sqlalchemy.orm.mapper import _mapper_registry

    for m in list(_mapper_registry):
        if len(m.primary_key) == 1 and m.primary_key[0].name == pkname:
            return m.class_

    raise RuntimeError('Unknown PK: %s' % pkname)


def changes_summary(changes):
    """Format a set of changes into a nice string.

    :param changes: a mapping of field names to ``(oldvalue, newvalue)`` tuples
    :rtype: a string

      >>> print(changes_summary(dict(a=(None, 1))))
      changed a to 1
      >>> print(changes_summary(dict(a=(False, True))))
      changed a from False to True
      >>> print(changes_summary(dict(a=(0,1), b=('foo','bar'))))
      changed a from 0 to 1 and b from "foo" to "bar"
      >>> print(changes_summary(dict(a=(0,1), b=(None,'bar'), c=(True,None))))
      changed a from 0 to 1, b to "bar" and c from True to None
    """

    summary = []
    for field in sorted(changes):
        oldvalue, newvalue = changes[field]

        if oldvalue is None or oldvalue == '':
            oldvalue = False
        else:
            if isinstance(oldvalue, str):
                oldvalue = '"%s"' % oldvalue
            else:
                oldvalue = str(oldvalue)

        if isinstance(newvalue, str):
            newvalue = '"%s"' % newvalue
        else:
            newvalue = str(newvalue)

        if oldvalue is False:
            summary.append('%s to %s' % (field, newvalue))
        else:
            summary.append('%s from %s to %s' % (field, oldvalue, newvalue))

    return 'changed ' + njoin(summary, localized=False)


def save_changes(sasess, request, modified, deleted, clogger=logger):
    """Save insertions, changes and deletions to the database.

    :param sasess: the SQLAlchemy session
    :param request: the Pyramid web request
    :param modified: a sequence of record changes, each represented by
                     a tuple of two items, the PK name and a
                     dictionary with the modified fields; if the value
                     of the PK field is null or 0 then the record is
                     considered new and will be inserted instead of updated
    :param deleted: a sequence of deletions, each represented by a tuple
                    of two items, the PK name and the ID of the record to
                    be removed
    :param clogger: where to log applied changes
    :rtype: a tuple of three lists, respectively inserted, modified and
            deleted record IDs, grouped in a dictionary keyed on PK name.
    """

    iids = []
    mids = []
    dids = []

    # Dictionary with last inserted PK ids: each newly inserted
    # primary key (records with id==0) is stored here by name, and
    # used for the homonym FK with value of 0. This let us insert a
    # new master record with its details in a single call.
    last_ins_ids = {}

    try:
        for key, mdata in modified:
            entity = entity_from_primary_key(key)

            fvalues = {}
            for f, v in mdata.items():
                if f != key:
                    if v != '':
                        fvalues[f] = v
                    else:
                        fvalues[f] = None

            # Update the NULL foreign keys with previously
            # inserted master ids
            for lik in last_ins_ids:
                if lik != key and fvalues.get(lik) == 0:
                    fvalues[lik] = last_ins_ids[lik]

            # If there are no changes, do not do anything
            if not fvalues:
                continue

            # If the PK is missing or None, assume it's a new record
            idrecord = int(mdata.get(key) or 0)

            if idrecord == 0:
                instance = entity(**fvalues)
                sasess.add(instance)
                sasess.flush()
                nextid = getattr(instance, key)
                iids.append({key: nextid})
                last_ins_ids[key] = nextid
                clogger.info('Inserted new %r', instance)
            else:
                instance = sasess.query(entity).get(idrecord)
                if instance is not None:
                    mids.append({key: idrecord})
                    changes = instance.update(fvalues)
                    sasess.flush()
                    clogger.info('Updated %r: %s', instance, changes_summary(changes))

        for key, ddata in deleted:
            entity = entity_from_primary_key(key)
            instance = sasess.query(entity).get(ddata)
            if instance is not None:
                sasess.delete(instance)
                dids.append({key: ddata})
                clogger.info('Deleted %r', instance)
    except SQLAlchemyError as e:
        clogger.warning('Changes rolled back due to an exception: %s', e)
        raise
    except:
        clogger.exception('Changes rolled back due to an exception')
        raise

    return iids, mids, dids


def dump(roots, specs):
    specs_by_qualname = {spec['entity']: spec for spec in specs}

    seen = set()
    entities = {}
    instances = {}

    def get_identity(instance):
        klass = instance.__class__
        qualname = instance.__module__ + '.' + klass.__name__
        if qualname not in specs_by_qualname:
            return

        spec = specs_by_qualname[qualname]

        key = spec['key']
        if callable(key):
            key = key(instance)
        if isinstance(key, str):
            identity = (qualname, getattr(instance, key))
        else:
            assert isinstance(key, (tuple, list))
            identity = (qualname, tuple(getattr(instance, f) for f in key))

        return identity

    def visit(instance):
        identity = get_identity(instance)
        if identity is None:
            return

        if identity in seen:
            return instances[identity]

        seen.add(identity)

        qualname = identity[0]
        spec = specs_by_qualname[qualname]
        key = spec['key']

        if qualname in entities:
            entity = entities[qualname]
        else:
            entity = entities[qualname] = {'entity': qualname, 'key': key, 'data': []}

        data = instances[identity] = {}
        entity['data'].append(data)

        attributes = []
        if isinstance(key, str):
            attributes.append(key)
        else:
            attributes.extend(key)

        other = spec.get('other', [])
        if isinstance(other, str):
            attributes.append(other)
        else:
            attributes.extend(other)

        for attr in attributes:
            value = getattr(instance, attr)

            if value is None:
                continue

            if isinstance(value, InstrumentedList):
                for i in value:
                    visit(i)
            else:
                try:
                    inspector = inspect(value)
                except NoInspectionAvailable:
                    data[attr] = value
                else:
                    if inspector.is_instance:
                        value = visit(value)
                        if value is not None:
                            data[attr] = value

        return data

    if isinstance(roots, (list, tuple)):
        for root in roots:
            visit(root)
    else:
        visit(roots)

    return [entities[spec['entity']] for spec in specs]



# -*- coding: utf-8 -*-
# :Project:   hurm -- Human Resources Manager
# :Created:   lun 14 dic 2015, 10.24.47, CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016 Lele Gaifax
#

from io import open
import os

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    README = f.read()
with open(os.path.join(here, 'CHANGES.rst'), encoding='utf-8') as f:
    CHANGES = f.read()
with open(os.path.join(here, 'version.txt'), encoding='utf-8') as f:
    VERSION = f.read().strip()

setup(
    name="hurm.db",
    version=VERSION,
    url="https://bitbucket.org/lele/hurm.db",

    description="Human Resources Manager, database definition",
    long_description=README + u'\n\n' + CHANGES,

    author="Lele Gaifax",
    author_email="lele@metapensiero.it",

    license="GPLv3+",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved ::"
        " GNU General Public License v3 or later (GPLv3+)",
        ],
    keywords='sqlalchemy postgresql',

    packages=['hurm.db'],
    namespace_packages=['hurm'],

    include_package_data=True,
    zip_safe=False,

    install_requires=[
        'cryptacular',
        'psycopg2',
        'setuptools',
        'sqlalchemy',
        'translationstring'
        ],
    extras_require={'dev': [
        'babel',
        'metapensiero.sphinx.patchdb',
        'metapensiero.sqlalchemy.dbloady',
        'metapensiero.tool.bump_version',
        'pyaxon',
        'readme',
    ]},
)



# -*- coding: utf-8 -*-
# :Project:   hurm -- SQLAlchemy declaration of table activities
# :Created:   mer 23 dic 2015 10:44:39 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2015, 2016 Lele Gaifax
#

import sqlalchemy as sa

from .. import metadata, translatable_string as _
from . import domains


activities = sa.Table(
    'activities', metadata,

    sa.Column(
        'idactivity', domains.integer_t, sa.Sequence('gen_idactivity', optional=True),
        nullable=False,
        primary_key=True,
        info={'label': _("Activity ID"),
              'hint': _("Unique ID of the activity")}),

    sa.Column(
        'description', domains.description_t,
        nullable=False,
        unique=True,
        info={'label': _("Activity"),
              'hint': _("Description of the activity")}),

    sa.Column(
        'note', domains.text_t,
        info={'label': _("Note"),
              'hint': _("Note")}),
)



# -*- coding: utf-8 -*-
# :Project:   hurm
# :Created:   ven 22 gen 2016 16:48:43 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from datetime import date, time

from sqlalchemy import exc as saexc

from hurm.db import entities


def test_out_edition_availability(session, edition_test, john_doe):
    na = entities.Availability()
    na.edition = edition_test
    na.person = john_doe
    na.date = date(2015, 5, 1)
    na.starttime = time(14, 0)
    na.endtime = time(16, 0)
    session.add(na)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'availability date outside allowed period' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()


def test_non_overlapping_availability(session, edition_test, john_doe):
    na = entities.Availability()
    na.edition = edition_test
    na.person = john_doe
    na.date = date(2016, 5, 1)
    na.starttime = time(15, 0)
    na.endtime = time(16, 0)
    session.add(na)
    try:
        session.flush()
    finally:
        session.rollback()


def test_overlapping_availability(session, edition_test, john_doe, bob_quartz):
    na = entities.Availability()
    na.edition = edition_test
    na.person = john_doe
    na.date = date(2016, 5, 1)
    na.starttime = time(14, 0)
    na.endtime = time(16, 0)
    session.add(na)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'overlapped availability' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()

    na = entities.Availability()
    na.edition = edition_test
    na.person = bob_quartz
    na.date = date(2016, 5, 1)
    na.starttime = time(8, 0)
    na.endtime = time(9, 0)
    session.add(na)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'overlapped availability' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()


def test_invalid_availability_edit(session, availability_john_doe_05_01):
    session.delete(availability_john_doe_05_01)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'duties outside person availability' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()

    availability_john_doe_05_01.endtime = time(10, 10)
    try:
        session.flush()
    except saexc.InternalError as e:
        assert 'duties outside person availability' in str(e)
    else:
        assert False, "Should raise an SQL error"
    finally:
        session.rollback()


def test_valid_availability_edit(session, availability_john_doe_04_30):
    availability_john_doe_04_30.note = 'Foo bar'
    try:
        session.flush()
    finally:
        session.rollback()

    availability_john_doe_04_30.starttime = time(10, 0)
    availability_john_doe_04_30.endtime = time(19, 0)
    try:
        session.flush()
    finally:
        session.rollback()
